from fqsdfqsdfqsdfqsd.tokens.NFT import NFT


class RT(NFT):
    _instances = set()

    def __init__(self, USDC_amount, wallet):
        super().__init__(USDC_amount, wallet)
        self._instances.add(self)

    @classmethod
    def get_instances(cls):
        return cls._instances

    @classmethod
    def del_instance(cls, instance):
        cls._instances.remove(instance)

    @classmethod
    def get_RT_in_circulation(cls):
        RTs = cls._instances
        total_amount = 0
        for rt in RTs:
            total_amount += rt.USDC_amount

        return total_amount


from fqsdfqsdfqsdfqsd.config import GlobalSettings
from fqsdfqsdfqsdfqsd.tokens import UT
from fqsdfqsdfqsdfqsd.contracts.repayment_contract import RepaymentContract
from fqsdfqsdfqsdfqsd.contracts.reserve_contract import ReserveContract
from dateutil.relativedelta import *

class DealContract:
    _instances = set()

    def __init__(self, reserve_contract, credix_wallet, principal=1_000_000, time_to_maturity=12, financing_fee=0.15, underwriter_fee=0.2, leverage_ratio=4):
        self.reserve_contract = reserve_contract
        self.credix_wallet = credix_wallet
        self.principal = int(principal)
        self.time_to_maturity = int(time_to_maturity) # in months
        self.financing_fee = financing_fee
        self.underwriter_fee = underwriter_fee
        self.leverage_ratio = int(leverage_ratio)

        self.go_live_date = None
        self.live = False
        self.junior_tranche_current = 0
        self.senior_tranche_open = False
        self.senior_tranche_current = 0
        self.USDC_balance = 0
        self.credit_outstanding = 0
        self.USDC_payout_investors = 0
        self.USDC_payout_underwriters = 0
        self.repayment_schedule = {}

        self._instances.add(self)

    @classmethod
    def get_instances(cls):
        return cls._instances

    @classmethod
    def del_instance(cls, instance):
        cls._instances.remove(instance)

    def set_repayment_schedule(self):
        for i in range(1, self.time_to_maturity + 1):
            repayment_date = GlobalSettings.CLOCK + relativedelta(months=+i)
            self.repayment_schedule[repayment_date.strftime("%Y-%m-%d")] = {
                "repaid": False,
                "principal": False
            }
            if i == self.time_to_maturity:
                self.repayment_schedule[repayment_date.strftime("%Y-%m-%d")] = {
                    "repaid": False,
                    "principal": True
                }

    def investor_interest(self):
        return self.financing_fee * (1 - self.underwriter_fee - GlobalSettings.BRIDGE_FINANCE_FEE)

    def underwriter_interest(self):
        return self.financing_fee * (1 - GlobalSettings.BRIDGE_FINANCE_FEE + self.leverage_ratio * self.underwriter_fee)

    def expected_yield_investors(self):
        return self.investor_interest() * self.senior_tranche_max()

    def expected_yield_underwriters(self):
        return self.underwriter_interest() * self.junior_tranche_max()

    def expected_yield_credix(self):
        return self.principal * self.financing_fee * GlobalSettings.BRIDGE_FINANCE_FEE

    def junior_tranche_max(self):
        return self.principal / (self.leverage_ratio + 1)

    def senior_tranche_max(self):
        return self.principal - self.junior_tranche_max()

    def expected_monthly_interest(self):
        return (self.principal * self.financing_fee) / self.time_to_maturity

    def fund_junior_tranche(self, underwriter_wallet, USDC_amount):
        junior_tranche_max = self.junior_tranche_max()
        if underwriter_wallet.USDC_balance < USDC_amount:
            print('not enough USDC in the underwriter wallet')
        elif self.junior_tranche_current >= junior_tranche_max:
            print("Junior tranche is fully funded")
        elif self.live:
            print("Deal is already live")
        else:
            if USDC_amount >= junior_tranche_max - self.junior_tranche_current:
                amount_to_fund = junior_tranche_max - self.junior_tranche_current
                self.senior_tranche_open = True
                print("Senior tranche is now open")
                print("Filling senior tranche")
                self.fund_senior_tranche()
            else:
                amount_to_fund = USDC_amount
            self.USDC_balance += amount_to_fund
            underwriter_wallet.USDC_balance -= amount_to_fund
            ut = UT(USDC_amount=amount_to_fund, wallet=underwriter_wallet, deal=self)
            underwriter_wallet.UT_balance.append(ut)
            self.junior_tranche_current += amount_to_fund

    def fund_senior_tranche(self):
        # first take funds from repayment pool
        amount_in_repayment_pool = RepaymentContract.USDC_amount
        amount_in_reserve_pool = ReserveContract.get_TV_reserve_pool()

        if self.senior_tranche_max() <= amount_in_repayment_pool + amount_in_reserve_pool:
            if amount_in_repayment_pool >= self.senior_tranche_max():
                RepaymentContract.USDC_amount -= self.senior_tranche_max()
            else:
                self.reserve_contract.swap_RT_for_IT(USDC_amount_needed=self.senior_tranche_max() - amount_in_repayment_pool,
                                                 deal=self)
                RepaymentContract.USDC_amount = 0

            self.go_live_date = GlobalSettings.CLOCK
            self.set_repayment_schedule()
            self.credit_outstanding = self.principal
            self.senior_tranche_current = 0
            self.junior_tranche_current = 0
            self.live = True
            print("The deal is now live")
        else:
            print("Not enough capital in reserve pool")

    def borrow_principal(self):
        self.USDC_balance = 0

    def repay_interest(self, interest_USDC_amount):
        print("-------------")
        print(f"repaying interest of {interest_USDC_amount} USDC")

        # pay out credix first
        credix_interest = interest_USDC_amount * GlobalSettings.BRIDGE_FINANCE_FEE
        self.credix_wallet.USDC_balance += credix_interest
        print(f"payout credix interest: {credix_interest} USDC")

        # pay out investors
        investor_interest = self.financing_fee * self.senior_tranche_max() * (1 - GlobalSettings.BRIDGE_FINANCE_FEE - self.underwriter_fee) / self.time_to_maturity
        if not interest_USDC_amount - investor_interest > 0:
            investor_interest = interest_USDC_amount
        RepaymentContract.USDC_amount += investor_interest
        self.USDC_payout_investors += investor_interest
        print(f"payout investors interest: {investor_interest} USDC")

        # pay out underwriters
        underwriters_interest = interest_USDC_amount - credix_interest - investor_interest
        self.USDC_payout_underwriters += underwriters_interest
        for ut in UT.get_instances().copy():
            if ut.deal == self:
                ut.wallet.USDC_balance += ut.USDC_amount / self.junior_tranche_max() * underwriters_interest

        print(f"payout underwriters interest: {underwriters_interest} USDC")
        print("-------------")

    def repay_principal(self, principal_USDC_amount):
        # pay out investors first
        payout_investors = self.senior_tranche_max() + self.expected_yield_investors() - self.USDC_payout_investors
        RepaymentContract.USDC_amount += payout_investors
        print("-------------")
        print("payout investors principal")
        print(payout_investors)

        # pay out underwriters second
        expected_payout_underwriters = self.junior_tranche_max() + self.expected_yield_underwriters() - self.USDC_payout_underwriters
        actual_payout_underwriters = 0
        if principal_USDC_amount - payout_investors >= expected_payout_underwriters:
            payout_underwriters_percentage = 1
        else:
            payout_underwriters_percentage = max(0, (principal_USDC_amount - payout_investors)/expected_payout_underwriters)

        for ut in UT.get_instances().copy():
            if ut.deal == self:
                USDC_yield = ut.USDC_amount * (1 + self.underwriter_interest()) * payout_underwriters_percentage
                ut.wallet.USDC_balance += USDC_yield
                actual_payout_underwriters += USDC_yield
                ut.wallet.del_UT(ut)
        print("-------------")
        self.credit_outstanding = 0
        self.live = False





from .reserve_contract import ReserveContract
from .repayment_contract import RepaymentContract
from .deal_contract import DealContract

__all__ = [
    "DealContract",
    "ReserveContract",
    "RepaymentContract"
]


from .investor_wallet import InvestorWallet
from .underwriter_wallet import UnderwriterWallet
from .credix_wallet import CredixWallet

__all__ = [
    "InvestorWallet",
    "UnderwriterWallet",
    "CredixWallet"
]


from fqsdfqsdfqsdfqsd.tokens import RT


class ReserveContract:
    def __init__(self, USDC_max = 10_000_000, compound_apy=0.018):
        self.USDC_max = USDC_max
        self.compound_apy = compound_apy

    def swap_RT_for_IT(self, USDC_amount_needed, deal):
        from fqsdfqsdfqsdfqsd.oracles import PricingOracle

        oldest_RTs = self.find_oldest_RTs()
        for rt in oldest_RTs:
            rt_USDC_amount = rt.USDC_amount
            USDC_amount_needed -= rt_USDC_amount
            if USDC_amount_needed > 0:
                rt.wallet.IT_balance += PricingOracle.USDC_to_IT(rt_USDC_amount)
                deal.senior_tranche_current += rt_USDC_amount
                deal.USDC_balance += rt_USDC_amount
                rt.wallet.RT_balance.remove(rt)
                RT.del_instance(rt)
            else:
                rt.wallet.IT_balance += PricingOracle.USDC_to_IT(USDC_amount_needed + rt_USDC_amount)
                deal.senior_tranche_current += USDC_amount_needed + rt_USDC_amount
                deal.USDC_balance += USDC_amount_needed + rt_USDC_amount
                rt.USDC_amount -= USDC_amount_needed + rt_USDC_amount
                break

    @classmethod
    def get_TV_reserve_pool(cls):
        TV = 0
        RTs = list(RT.get_instances())
        for rt in RTs:
            TV += rt.USDC_amount

        return TV

    @staticmethod
    def fund(USDC_amount, investor_wallet):
        if investor_wallet.USDC_balance < USDC_amount:
            pass
        else:
            # print(f"Investor received RTs for {USDC_amount} USDC")
            reserve_token = RT(USDC_amount, investor_wallet)
            investor_wallet.RT_balance.append(reserve_token)
            investor_wallet.USDC_balance -= USDC_amount

    @staticmethod
    def withdraw(rt, investor_wallet):
        USDC_amount = rt.USDC_amount
        investor_wallet.USDC_balance += USDC_amount
        investor_wallet.RT_balance.remove(rt)

    @staticmethod
    def find_oldest_RTs():
        all_RTs = list(RT.get_instances())

        return sorted(all_RTs, key=lambda x: x.issuance_date)

    # TODO implement compound interest model
    @staticmethod
    def return_daily_interest(self):
        daily_apy = self.compound_apy / 365
        for rt in RT.get_instances():
            accrued_interest = rt.USDC_amount * daily_apy
            rt.wallet.USDC_balance += accrued_interest


from fqsdfqsdfqsdfqsd.contracts import RepaymentContract
from fqsdfqsdfqsdfqsd.wallets import InvestorWallet
from fqsdfqsdfqsdfqsd.contracts import DealContract

class PricingOracle:
    probability_of_default = 0.04
    loss_given_default = 0.5
    discount_rate = 0.05
    n_seconds_year = 31536000

    @classmethod
    def IT_price(cls):
        USDC_repayed = RepaymentContract.USDC_amount
        credit_outstanding_senior = cls.get_credit_outstanding_senior()
        USDC_deals_senior = cls.get_USDC_deals_senior()
        IT_in_circulation = cls.get_IT_in_circulation()

        if IT_in_circulation == 0:
            price = 1
        else:
            price = (USDC_repayed + USDC_deals_senior + credit_outstanding_senior) / IT_in_circulation
            # print("----")
            # print(f"USDC_repayed: {USDC_repayed}")
            # print(f"USDC_deals_senior: {USDC_deals_senior}")
            # print(f"credit_outstanding_senior: {credit_outstanding_senior}")
            # print(f"IT_in_circulation: {IT_in_circulation}")
            # print(f"price: {price}")
            # print("---")

        return price

    @classmethod
    def get_USDC_deals_senior(cls):
        USDC_deals_senior = 0
        deals = DealContract.get_instances()
        for deal in deals:
            USDC_deals_senior += deal.senior_tranche_current

        return USDC_deals_senior

    @classmethod
    def get_credit_outstanding_senior(cls):
        credit_outstanding_senior = 0
        deals = DealContract.get_instances()
        for deal in deals:
            credit_outstanding_senior += deal.credit_outstanding * (deal.leverage_ratio / (deal.leverage_ratio + 1))

        return credit_outstanding_senior

    @classmethod
    def USDC_to_IT(cls, USDC_amount):
        return USDC_amount / cls.IT_price()

    @classmethod
    def IT_to_USDC(cls, IT_amount):
        return IT_amount * cls.IT_price()

    @classmethod
    def get_IT_in_circulation(cls):
        IT_in_circulation = 0
        investor_wallets = InvestorWallet.get_instances()
        for investor_wallet in investor_wallets:
            IT_in_circulation += investor_wallet.IT_balance

        return IT_in_circulation

    # @classmethod
    # def PV_all_deals(cls):
    #     PV_all_deals = 0
    #     deals = DealContract.get_instances()
    #     for deal in deals:
    #         if deal.live:
    #             PV_all_deals += cls.PV_one_deal(deal)
    #
    #     return PV_all_deals
    #
    # @classmethod
    # def PV_one_deal(cls, deal):
    #     n_months_till_maturity = cls.diff_month((deal.go_live_date + relativedelta(months=deal.time_to_maturity)), GlobalSettings.CLOCK)
    #     n_seconds_till_maturity = ((deal.go_live_date + relativedelta(months=deal.time_to_maturity)) - GlobalSettings.CLOCK).total_seconds()
    #     payment_for_period = deal.financing_fee * deal.principal / deal.time_to_maturity
    #     future_CF = deal.principal + payment_for_period * n_months_till_maturity
    #     expected_loss = future_CF * (n_months_till_maturity / deal.time_to_maturity) * cls.probability_of_default * cls.loss_given_default
    #     risk_adjusted_future_CF = future_CF - expected_loss
    #     PV = risk_adjusted_future_CF / (1 + cls.discount_rate) ** (n_seconds_till_maturity / cls.n_seconds_year)
    #
    #     if PV < 0:
    #         return 0
    #     else:
    #         return PV
    #
    # @staticmethod
    # def diff_month(d1, d2):
    #     return (d1.year - d2.year) * 12 + d1.month - d2.month


from fqsdfqsdfqsdfqsd.config import GlobalSettings

class NFT:
    def __init__(self, USDC_amount, wallet):
        self.USDC_amount = USDC_amount
        self.issuance_date = GlobalSettings.CLOCK
        self.wallet = wallet


from fqsdfqsdfqsdfqsd.tokens.NFT import NFT


class UT(NFT):
    _instances = set()

    def __init__(self, USDC_amount, wallet, deal):
        super().__init__(USDC_amount, wallet)
        self.deal = deal
        self._instances.add(self)

    @classmethod
    def get_instances(cls):
        return cls._instances

    @classmethod
    def del_instance(cls, instance):
        cls._instances.remove(instance)

    @classmethod
    def get_UT_in_circulation(cls):
        UTs = cls._instances
        total_amount = 0
        for ut in UTs:
            total_amount += ut.USDC_amount

        return total_amount



from fqsdfqsdfqsdfqsd.wallets.wallet import Wallet


class CredixWallet(Wallet):
    _instances = set()

    def __init__(self, USDC_balance=0):
        super().__init__(USDC_balance=USDC_balance)
        self.UT_balance = None
        self.RT_balance = None
        self.IT_balance = None
        self._instances.add(self)

    @classmethod
    def get_instances(cls):
        return cls._instances

    @classmethod
    def del_instance(cls, instance):
        cls._instances.remove(instance)


from fqsdfqsdfqsdfqsd.wallets import UnderwriterWallet, InvestorWallet, CredixWallet
from fqsdfqsdfqsdfqsd.tokens import UT, RT
from fqsdfqsdfqsdfqsd.contracts import ReserveContract, RepaymentContract, DealContract
from fqsdfqsdfqsdfqsd.config import GlobalSettings
from fqsdfqsdfqsdfqsd.oracles import PricingOracle

import collections
import random
from datetime import datetime, timedelta
from dateutil.relativedelta import *
import pandas as pd

DEFAULT_CONFIG = {
    "underwriters": {
        "amount": 10,
        "USDC_balance": [100,500]
    },
    "investors": {
        "amount": 50,
        "USDC_balance": [40,50]
    },
    "deals": [
        {
            "months_after_sim_start": 0,
            "time_to_maturity": 12,
            "principal": 1000,
            "financing_fee": 0.15,
            "underwriter_fee": 0.2,
            "leverage_ratio": 4
        },
        {
            "months_after_sim_start": 3,
            "time_to_maturity": 6,
            "principal": 2000,
            "financing_fee": 0.2,
            "underwriter_fee": 0.2,
            "leverage_ratio": 4
        }
    ],
    "simulation": {
        "start_date": "2021-01-01",
        "duration_months": 20,
    }
}


class MainSimulation:
    def __init__(self, config=DEFAULT_CONFIG):
        self.config = self.update_config(config)
        self.credix_wallet = CredixWallet()
        self.reserve = ReserveContract()

    def run(self):
        self.reset_clock()
        self.reset_instances()
        self.initialize_actors()
        dates = []
        IT_prices = []
        RT_in_circulation = []
        IT_in_circulation = []
        UT_in_circulation = []
        deal_go_live = []
        repayment_pool = []
        total_seconds = ((GlobalSettings.CLOCK + relativedelta(months=int(self.config["simulation"]["duration_months"]))) - GlobalSettings.CLOCK).total_seconds()
        total_hours = int(total_seconds/3600)
        total_days = int(total_seconds/(3600*24))
        date_str_old = "abc"
        # current_month = 13
        for i in range(0, total_days):
            date_str = GlobalSettings.CLOCK.strftime("%Y-%m-%d")
            if date_str != date_str_old:
                # make repayments for deals
                for deal in DealContract.get_instances():
                    if deal.live:
                        self.make_repayment(deal, date_str)

                # launch and fund deal
                for deal_config in self.config["deals"]:
                    # deal_go_live_date = datetime.strptime(self.config["simulation"]["start_date"],
                    #                                       "%Y-%m-%d") + relativedelta(
                    #     months=deal_config["months_after_sim_start"])
                    if deal_config["deal_go_live"] == date_str:
                        print("-----------------")
                        print("Initializing deal")
                        print("-----------------")
                        deal = self.initialize_deal(deal_config)
                        if deal.live:
                            deal_go_live.append(GlobalSettings.CLOCK.strftime("%Y-%m-%d"))

                date_str_old = date_str

            # if GlobalSettings.CLOCK.month != current_month:
            dates.append(GlobalSettings.CLOCK.strftime("%Y-%m-%d"))
            IT_prices.append(PricingOracle.IT_price())
            RT_in_circulation.append(RT.get_RT_in_circulation())
            IT_in_circulation.append(PricingOracle.get_IT_in_circulation())
            UT_in_circulation.append(UT.get_UT_in_circulation())
            repayment_pool.append(RepaymentContract.USDC_amount)
                #current_month = GlobalSettings.CLOCK.month

            GlobalSettings.CLOCK += timedelta(seconds=3600 * 24)

        simulation_df = pd.DataFrame.from_dict({"date": dates, "IT price": IT_prices, "RT": RT_in_circulation, "IT": IT_in_circulation,
                                "UT": UT_in_circulation, "repayment pool": repayment_pool})

        simulation_df["APY 30d trailing"] = self.calculate_trailing_APY(simulation_df[["IT price"]])

        return simulation_df, deal_go_live

    def reset_clock(self):
        GlobalSettings.CLOCK = datetime.strptime(self.config["simulation"]["start_date"], "%Y-%m-%d")

    @staticmethod
    def calculate_trailing_APY(price_df):
        price_df['apy'] = ((((1 + price_df['IT price'] - price_df.shift(30)['IT price']).pow(1/30.0)).pow(365.0))-1)*100
        return price_df['apy']


    @staticmethod
    def reset_instances():
        for deal in DealContract.get_instances().copy():
            DealContract.del_instance(deal)
        for wallet in InvestorWallet.get_instances().copy():
            InvestorWallet.del_instance(wallet)
        for wallet in UnderwriterWallet.get_instances().copy():
            UnderwriterWallet.del_instance(wallet)
        for wallet in CredixWallet.get_instances().copy():
            CredixWallet.del_instance(wallet)
        for rt in RT.get_instances().copy():
            RT.del_instance(rt)
        for ut in UT.get_instances().copy():
            UT.del_instance(ut)
        RepaymentContract.USDC_amount = 0


    @staticmethod
    def make_repayment(deal, date_str):
        if date_str in deal.repayment_schedule:
            interest_amount = deal.principal * deal.financing_fee / deal.time_to_maturity
            if not deal.repayment_schedule[date_str]["repaid"] and not deal.repayment_schedule[date_str]["principal"]:
                deal.repayment_schedule[date_str]["repaid"] = True
                deal.repay_interest(interest_amount)
            elif not deal.repayment_schedule[date_str]["repaid"] and deal.repayment_schedule[date_str]["principal"]:
                deal.repayment_schedule[date_str]["repaid"] = True
                deal.repay_interest(interest_amount)
                deal.repay_principal(deal.principal)
        else:
            pass

    def initialize_actors(self):
        random.seed(123)
        self.initialize_investors()
        self.initialize_underwriters()

    def initialize_investors(self):
        for i in range(0, self.config["investors"]["amount"]):
            min_USDC_balance = self.config["investors"]["USDC_balance"][0]
            max_USDC_balance = self.config["investors"]["USDC_balance"][1]
            InvestorWallet(USDC_balance=random.randint(min_USDC_balance, max_USDC_balance))

        for investor_wallet in InvestorWallet.get_instances():
            self.reserve.fund(USDC_amount=investor_wallet.USDC_balance, investor_wallet=investor_wallet)

    def initialize_underwriters(self):
        for i in range(0, self.config["underwriters"]["amount"]):
            min_USDC_balance = self.config["underwriters"]["USDC_balance"][0]
            max_USDC_balance = self.config["underwriters"]["USDC_balance"][1]
            UnderwriterWallet(USDC_balance=random.randint(min_USDC_balance, max_USDC_balance))

    def initialize_deal(self, deal_config):
            deal_config_new = deal_config.copy()
            deal_config_new.pop("deal_go_live")
            deal = DealContract(**deal_config_new, reserve_contract=self.reserve, credix_wallet=self.credix_wallet)
            for underwriter_wallet in UnderwriterWallet.get_instances():
                if not deal.senior_tranche_open:
                    deal.fund_junior_tranche(underwriter_wallet=underwriter_wallet, USDC_amount=int(deal.principal/4))
            return deal

    def update_config(self, config):
        return self.update_dict(DEFAULT_CONFIG, config)

    def update_dict(self, d, u):
        for k, v in u.items():
            if isinstance(d, collections.Mapping):
                if isinstance(v, collections.Mapping):
                    r = self.update_dict(d.get(k, {}), v)
                    d[k] = r
                else:
                    d[k] = u[k]
            else:
                d = {k: u[k]}
        return d



import uuid
from fqsdfqsdfqsdfqsd.tokens import UT, RT

class Wallet:
    def __init__(self, USDC_balance=1000):
        self.wallet_address = str(uuid.uuid4())
        self.USDC_balance = USDC_balance
        self.UT_balance = list()
        self.RT_balance = list()
        self.IT_balance = 0

    def del_UT(self, instance):
        try:
            self.UT_balance.remove(instance)
            UT.del_instance(instance)
        except:
            pass

    def del_RT(self, instance):
        try:
            self.RT_balance.remove(instance)
            RT.del_instance(instance)
        except:
            pass


from setuptools import setup, find_packages

setup(name='fqsdfqsdfqsdfqsd',
      version='0.11',
      description='fqsdfqsdfqsdfqsd',
      author='fqsdfqsdfqsdfqsd',
      author_email='fqsdfqsdfqsdfqsdn@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
              'python-dateutil',
              'pandas'
          ],
      zip_safe=False)